@@ -1,5 +1,5 @@
 /* Test suite for exclude.
-   Copyright (C) 2009-2015 Free Software Foundation, Inc.
+   Copyright (C) 2009-2016 Free Software Foundation, Inc.
    This file is part of the GNUlib Library.
 
    This program is free software: you can redistribute it and/or modify
@@ -24,7 +24,6 @@
 #include <fnmatch.h>
 
 #include "exclude.h"
-#include "progname.h"
 #include "error.h"
 #include "argmatch.h"
 
@@ -75,8 +74,6 @@ main (int argc, char **argv)
   int exclude_options = 0;
   struct exclude *exclude = new_exclude ();
 
-  set_program_name (argv[0]);
-
   if (argc == 1)
     error (1, 0, "usage: %s file -- words...", argv[0]);
 
