@@ -29,7 +29,6 @@ SIGNATURE_CHECK (openat, int, (int, char const *, int, ...));
 #include <stdio.h>
 #include <unistd.h>
 
-#include "progname.h"
 #include "macros.h"
 
 #define BASE "test-openat.t"
@@ -63,8 +62,6 @@ main (int argc _GL_UNUSED, char *argv[])
 {
   int result;
 
-  set_program_name (argv[0]);
-
   /* Test behaviour for invalid file descriptors.  */
   {
     errno = 0;
