@@ -20,6 +20,19 @@
 
 #include <config.h>
 #include "search.h"
+#include "verify.h"
+
+/* Wrap a fatal 3-argument use of "error" (with literal nonzero exit
+   status), so that static analyzers like clang-analyzer and GCC 7's
+   -Wimplicit-fallthrough know this "function" does not return.  */
+#define die(st, fmt, arg)						\
+  do									\
+    {									\
+      verify ((st) != 0);						\
+      error (0, fmt, arg);						\
+      exit (st);							\
+    }									\
+  while (0)
 
 #if HAVE_LIBPCRE
 # include <pcre.h>
@@ -114,7 +127,7 @@ Pcompile (char const *pattern, size_t size)
 
   if (1 < MB_CUR_MAX)
     {
-      if (! using_utf8 ())
+      if (! localeinfo.using_utf8)
         error (EXIT_TROUBLE, 0,
                _("-P supports only unibyte and UTF-8 locales"));
       multibyte_locale = true;
@@ -254,7 +267,7 @@ Pexecute (char *buf, size_t size, size_t *match_size,
           /* Skip past bytes that are easily determined to be encoding
              errors, treating them as data that cannot match.  This is
              faster than having pcre_exec check them.  */
-          while (mbclen_cache[to_uchar (*p)] == (size_t) -1)
+          while (localeinfo.sbclen[to_uchar (*p)] == -1)
             {
               p++;
               subject = p;
@@ -337,15 +350,15 @@ Pexecute (char *buf, size_t size, size_t *match_size,
           break;
 
         case PCRE_ERROR_NOMEMORY:
-          error (EXIT_TROUBLE, 0, _("memory exhausted"));
+          die (EXIT_TROUBLE, 0, _("memory exhausted"));
 
 # if PCRE_STUDY_JIT_COMPILE
         case PCRE_ERROR_JIT_STACKLIMIT:
-          error (EXIT_TROUBLE, 0, _("exhausted PCRE JIT stack"));
+          die (EXIT_TROUBLE, 0, _("exhausted PCRE JIT stack"));
 # endif
 
         case PCRE_ERROR_MATCHLIMIT:
-          error (EXIT_TROUBLE, 0, _("exceeded PCRE's backtracking limit"));
+          die (EXIT_TROUBLE, 0, _("exceeded PCRE's backtracking limit"));
 
         default:
           /* For now, we lump all remaining PCRE failures into this basket.
